""""Package metadata."""

__title__ = "automathon"
__description__ = "A Python library for simulating and visualizing finite automata"
__url__ = "https://github.com/rohaquinlop/automathon.git"
__version__ = "0.0.13"
__author__ = "Robin Hafid Quintero Lopez"
__author_email__ = "rohaquinlop301@gmail.com"
